#ifndef _GLERROR_H
#define _GLERROR_H

#include <stdio.h>
#ifdef _WIN32
#include "windows.h"
static char errormessage[512];
#endif

#ifdef _DEBUG
#ifdef _WIN32
#define GLERROR1( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_INVALID_ENUM -- GLenum argument out of range\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#define GLERROR2( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_INVALID_VALUE -- Numeric argument out of range\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#define GLERROR3( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_INVALID_OPERATION -- Operation illegal in current state\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#define GLERROR4( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_STACK_OVERFLOW -- Command would cause stack overflow\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#define GLERROR5( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_STACK_UNDERFLOW -- Command would cause stack underflow\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#define GLERROR6( locString ) \
        sprintf( errormessage, "%s --- OpenGL Error: GL_OUT_OF_MEMORY -- Not enough memory left to execute command\n", locString ); \
        MessageBox( 0, errormessage, "CheckGLError()", MB_OK ); \
        fprintf( stderr, "%s", errormessage );
#else
#define GLERROR1( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_INVALID_ENUM -- GLenum argument out of range\n", locString );
#define GLERROR2( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_INVALID_VALUE -- Numeric argument out of range\n", locString );
#define GLERROR3( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_INVALID_OPERATION -- Operation illegal in current state\n", locString );
#define GLERROR4( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_STACK_OVERFLOW -- Command would cause stack overflow\n", locString );
#define GLERROR5( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_STACK_UNDERFLOW -- Command would cause stack underflow\n", locString );
#define GLERROR6( locString ) \
        fprintf( stderr, "%s --- OpenGL Error: GL_OUT_OF_MEMORY -- Not enough memory left to execute command\n", locString );
#endif

#define CheckGLError( locString ) \
    { \
        GLenum errCode; \
        if ((errCode = glGetError()) != GL_NO_ERROR) \
        { \
            switch( errCode ) \
            { \
                case GL_INVALID_ENUM: \
                    GLERROR1( locString ) \
                    break; \
                case GL_INVALID_VALUE: \
                    GLERROR2( locString ) \
                    break; \
                case GL_INVALID_OPERATION: \
                    GLERROR3( locString ) \
                    break; \
                case GL_STACK_OVERFLOW: \
                    GLERROR4( locString ) \
                    break; \
                case GL_STACK_UNDERFLOW: \
                    GLERROR5( locString ) \
                    break; \
                case GL_OUT_OF_MEMORY: \
                    GLERROR6( locString ) \
                    break; \
            } \
        } \
    }
#else
#define CheckGLError( locString )
#endif

#endif
