#include "shape.h"

#include <float.h>
#include <math.h>
#include <string.h>
#include <GLES/gl.h>
#include "nvfxmath/nvfxmath.h"
#include "nvtexture/nvtexture.h"
#include "nvmediapath/nvmediapath.h"
#include "glerror.h"

#define M_PI       3.14159265358979323846

/* flag indicating if lod_clamping is supported. */
extern int lod_clamp_supported;

static GLuint texidA[4] = { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF };
static GLuint texidB[4] = { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF };
static int current = 2;


#define DOT_SF16( a, b ) \
    NvSFxAdd( \
		NvSFxAdd( \
		    NvSFxMul( a[0], b[0] ), \
			NvSFxMul( a[1], b[1] ) \
		), \
		NvSFxMul( a[2], b[2] ) \
	)

static void ShapeInitializeTextures()
{
    char filename[64];
    char pathname[128];
    char fullpathname[256];
    char *names[4] = { "nvlobby", "patio", "nvamphi", "nvwalkway" };
    int i;
    int j;

    /* Only initialize the textures the first time. */
    if ( texidA[0] != 0xFFFFFFFF )
        return;

    glGenTextures( 4, texidA );
    glGenTextures( 4, texidB );

    for ( i = 0; i < 4; i++ )
    {
        for ( j = 0; j < 2; j++ )
        {
            if ( j == 0 )
            {
                sprintf( filename, "%sA.dds", names[i] );
                glBindTexture( GL_TEXTURE_2D, texidA[i] );
            }
            else
            {
                sprintf( filename, "%sB.dds", names[i] );
                glBindTexture( GL_TEXTURE_2D, texidB[i] );
            }
            sprintf( pathname, "MEDIA/textures/paraboloidmaps/%s/", names[i] );
            LocateMediaFile( filename, pathname, fullpathname );
            if ( !LoadTextureDDS( fullpathname ) )
            {
                exit(-1);
            }
            glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
            glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
            glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
            glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
            if ( lod_clamp_supported )
                glTexParameterf( GL_TEXTURE_2D, /* GL_TEXTURE_MAX_LOD_SGIS */ 0x813B, 4.0f );
        }
    }

    CheckGLError( "ShapeInitializeTextures" );
}

void ShapeInitialize( Shape *shape )
{
    shape->final_drag = NvSFxFloat2Fixed( 0.99f );
    shape->initial_drag = NvSFxFloat2Fixed( 0.7f );
    shape->drag = NvSFxFloat2Fixed( 0.7f );

    ShapeInitializeTextures();
}

void ShapeCycleNextCubeMap( Shape *shape )
{
    current = (current+1) % 4;
}

void ShapeDraw( Shape *shape )
{
    int i;
    int triCount  = 0;
    int vertCount = 0;
    Tristrip *tstrip = shape->tristrips;

    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FIXED, sizeof(Vertex), shape->vertices[0].p );

    /* Begin 1st unit coords and setup. */
    glActiveTexture( GL_TEXTURE0 );
    glClientActiveTexture( GL_TEXTURE0 );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );
    glEnable( GL_TEXTURE_2D );
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);

    glTexCoordPointer( 2, GL_FIXED, sizeof(Vertex), shape->vertices[0].tcmapA );
    glBindTexture( GL_TEXTURE_2D, texidA[current] );

    glEnableClientState( GL_COLOR_ARRAY );
    glColorPointer( 4, GL_UNSIGNED_BYTE, sizeof(Vertex), &(shape->vertices[0].alpha) );

    glEnable( GL_ALPHA_TEST );
    glAlphaFunc( GL_LEQUAL, 0.5f );

    glEnable( GL_CULL_FACE );
    glCullFace( GL_FRONT );

#if 1
    for ( i = 0; i < shape->numTristrips; i++, tstrip++ )
    {
        glDrawElements( GL_TRIANGLE_STRIP,
                        tstrip->numVerts,
                        GL_UNSIGNED_SHORT,
                        tstrip->indices );

        triCount  += tstrip->numVerts-2;
        vertCount += tstrip->numVerts;
    }
#endif

    glTexCoordPointer( 2, GL_FIXED, sizeof(Vertex), shape->vertices[0].tcmapB );
    glBindTexture( GL_TEXTURE_2D, texidB[current] );

    glAlphaFunc( GL_GREATER, 0.5f );

    triCount  = 0;
    vertCount = 0;
    tstrip = shape->tristrips;

#if 1
    for ( i = 0; i < shape->numTristrips; i++, tstrip++ )
    {
        glDrawElements( GL_TRIANGLE_STRIP,
                        tstrip->numVerts,
                        GL_UNSIGNED_SHORT,
                        tstrip->indices );

        triCount  += tstrip->numVerts-2;
        vertCount += tstrip->numVerts;
    }
#endif

    glDisable( GL_CULL_FACE );

    glDisable( GL_ALPHA_TEST );
    glDisableClientState( GL_COLOR_ARRAY );

    glActiveTexture( GL_TEXTURE0 );
    glClientActiveTexture( GL_TEXTURE0 );
    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisable( GL_TEXTURE_2D );

    glDisableClientState ( GL_VERTEX_ARRAY );
}

void ShapeDrawEdges( Shape *shape )
{
    int i, j;
    Tristrip *tstrip = shape->tristrips;
    unsigned short *edgeIndices;

    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
    glEnableClientState ( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FIXED, sizeof(Vertex), shape->vertices[0].p );

    edgeIndices = (unsigned short*)
                  malloc(((shape->tristrips->numVerts+1)>>1) * sizeof(unsigned short));

    for ( i = 0; i < shape->numTristrips; i++, tstrip++ )
    {
        // draw the zig-zags
        glDrawElements( GL_LINE_STRIP,
                        tstrip->numVerts,
                        GL_UNSIGNED_SHORT,
                        tstrip->indices );

        // draw one "side" of the tristrip (don't need to draw the other
        // side because our neighboring tristrip will do it)
        for ( j = 0; j < tstrip->numVerts; j += 2 )
        {
            edgeIndices[j>>1] = tstrip->indices[j];
        }
        glDrawElements( GL_LINE_STRIP,
                        (tstrip->numVerts+1)>>1,
                        GL_UNSIGNED_SHORT,
                        edgeIndices );
    }

    free(edgeIndices);
}

void ShapeDrawVertices( Shape *shape )
{
    int i, j;
    Tristrip *tstrip = shape->tristrips;
    unsigned short *edgeIndices;

    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
    glEnableClientState ( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FIXED, sizeof(Vertex), shape->vertices[0].p );

    edgeIndices = (unsigned short*)
                  malloc(((shape->tristrips->numVerts+1)>>1) * sizeof(unsigned short));

    for ( i = 0; i < shape->numTristrips; i++, tstrip++ )
    {
        // draw one "side" of the tristrip (don't need to draw the other
        // side because our neighboring tristrip will do it)
        for ( j = 0; j < tstrip->numVerts; j += 2 )
        {
            edgeIndices[j>>1] = tstrip->indices[j];
        }
        glDrawElements( GL_POINTS,
                        (tstrip->numVerts+1)>>1,
                        GL_UNSIGNED_SHORT,
                        edgeIndices );
    }

    free( edgeIndices );
}


void ShapeCalcNormals( Shape *shape )
{
    int i, j;

    Vertex **verts;
    Tristrip *tstrip;
    Vertex *v0, *v1, *v2;
    NvSFx ax, ay, az;
    NvSFx bx, by, bz;
    NvSFx nx, ny, nz;
    NvSFx sign;

    for ( i = 0, tstrip = shape->tristrips; i < shape->numTristrips; i++, tstrip++ )
    {

		/*
        Vertex *v0, *v1, *v2;
        fix ax, ay, az;
        fix bx, by, bz;
        fix sign = 1.0f;
		*/

		sign = 0x00010000;
        verts = tstrip->vertices;

        v1 = verts[0];
        v2 = verts[1];

		/*
        ax = v2->p[0] - v1->p[0];
        ay = v2->p[1] - v1->p[1];
        az = v2->p[2] - v1->p[2];
		*/

        ax = NvSFxSub( v2->p[0], v1->p[0] );
        ay = NvSFxSub( v2->p[1], v1->p[1] );
        az = NvSFxSub( v2->p[2], v1->p[2] );

		/*
        for ( j = 0; j < tstrip->numVerts-2; j++, verts++, sign *= -1.0f )
		*/
        for ( j = 0; j < tstrip->numVerts-2; j++, verts++ )
        {
            v0 = v1;
            v1 = v2;
            v2 = verts[2];

            // Copy over the previous vector.  We invert the direction
            // every other vertex to ping-pong the normal.
			/*
            bx = sign * ax;
            by = sign * ay;
            bz = sign * az;

            ax = v2->p[0] - v1->p[0];
            ay = v2->p[1] - v1->p[1];
            az = v2->p[2] - v1->p[2];

            nx = ay*bz - az*by;
            ny = az*bx - ax*bz;
            nz = ax*by - ay*bx;

            v0->n[0] += nx;
            v0->n[1] += ny;
            v0->n[2] += nz;

            v1->n[0] += nx;
            v1->n[1] += ny;
            v1->n[2] += nz;

            v2->n[0] += nx;
            v2->n[1] += ny;
            v2->n[2] += nz;
			*/

            bx = NvSFxMul( sign, ax );
            by = NvSFxMul( sign, ay );
            bz = NvSFxMul( sign, az );

            ax = NvSFxSub( v2->p[0], v1->p[0] );
            ay = NvSFxSub( v2->p[1], v1->p[1] );
            az = NvSFxSub( v2->p[2], v1->p[2] );

            nx = NvSFxSub( NvSFxMul( ay, bz ), NvSFxMul( az, by ) );
            ny = NvSFxSub( NvSFxMul( az, bx ), NvSFxMul( ax, bz ) );
            nz = NvSFxSub( NvSFxMul( ax, by ), NvSFxMul( ay, bx ) );

			v0->n[0] = NvSFxAdd( v0->n[0], nx );
			v0->n[1] = NvSFxAdd( v0->n[1], ny );
			v0->n[2] = NvSFxAdd( v0->n[2], nz );

			v1->n[0] = NvSFxAdd( v1->n[0], nx );
			v1->n[1] = NvSFxAdd( v1->n[1], ny );
			v1->n[2] = NvSFxAdd( v1->n[2], nz );

			v2->n[0] = NvSFxAdd( v2->n[0], nx );
			v2->n[1] = NvSFxAdd( v2->n[1], ny );
			v2->n[2] = NvSFxAdd( v2->n[2], nz );

			if ( sign == 0x00010000 )
				sign = 0xFFFF0000;
			else
				sign = 0x00010000;
        }
    }
}

/*
 *
   Performs dual-paraboloid texture coordinate generation for every vertex in the shape.

   Given:
      eye -- position of the eye
      v   -- vertex position
      n   -- vertex normal

   Assumes:
      All input data (eye, v, and n) are in object space.

   Computes:
      Two sets of texture coordinates (s,t) based on a reflection vector, r.

      front map's coordinates
         s = 0.5 * (rx/(1-rz)) + 0.5
         t = 0.5 * (ry/(1-rz)) + 0.5

      back map's coordinates
         s = 0.5 * (-rx/(rz+1)) + 0.5
         t = 0.5 * (-ry/(rz+1)) + 0.5
      
      where
         r = u - 2*dot(u,n)*n
         u = v - eye

      Also computes an alpha value based on the z component of the reflection vector.
         alpha = 0.5 * rz + 0.5

   Side effects:
      The generated per-vertex texture coordinates (s,t) and alpha values will overwrite
      any pre-existing data.

   Additional Notes:
      For additional background information on dual-paraboloid mapping and a derivation
      of the above formulas, see the notes located here:

      http://developer.nvidia.com/object/dual_paraboloid_env_mapping.html      

      The inverse mapping (which is useful for creating paraboloid maps) is given by:

      front
        rx = 2 s / (s^2 + t^2 + 1)                  
        ry = 2 t / (s^2 + t^2 + 1)
        rz = (-1 + s^2 + t^2) / (s^2 + t^2 + 1)

      back
        rx = - 2 s / (s^2 + t^2 + 1)
        ry = - 2 t / (s^2 + t^2 + 1)
        rz = - (-1 + s^2 + t^2) / (s^2 + t^2 + 1
 */

void CalcTexcoordsFromNormal( Shape *shape, NvSFx eye[3] )
{
    int i;
    NvSFx *vobj;
    NvSFx *nobj;
    NvSFx n[3];
    NvSFx u[3];
    NvSFx r[3];
    NvSFx udotn;
    NvSFx m;
    NvSFx *tc;
    NvSFx len;

    for ( i = 0; i < shape->numVerts; i++ )
    {
        vobj = shape->vertices[i].p;
        nobj = shape->vertices[i].n;

		/*
        len = 1.0f / sqrt( float(nobj[0]* nobj[0] + nobj[1]*nobj[1] + nobj[2]*nobj[2]) );
		*/
		len = DOT_SF16( nobj, nobj );
        len = NvSFxSqrt( len );
		if ( len != 0x00000000 )
			len = NvSFxRecip( len );

		/*
        v[0] = vobj[0] + eye[0];
        v[1] = vobj[1] + eye[1];
        v[2] = vobj[2] + eye[2];
        n[0] = len*nobj[0];
        n[1] = len*nobj[1];
        n[2] = len*nobj[2];
		*/

        n[0] = NvSFxMul( len, nobj[0] );
        n[1] = NvSFxMul( len, nobj[1] );
        n[2] = NvSFxMul( len, nobj[2] );
        u[0] = NvSFxAdd( vobj[0], eye[0] );
        u[1] = NvSFxAdd( vobj[1], eye[1] );
        u[2] = NvSFxAdd( vobj[2], eye[2] );

        /* compute the unit length viewing vector */
		/*
        u[0] = v[0] - 0.0f;
        u[1] = v[1] - 0.0f;
        u[2] = v[2] - 0.0f;
        udotn = 1.0f / sqrt( float(u[0]*u[0] + u[1]*u[1] + u[2]*u[2]) );
        u[0] *= udotn;
        u[1] *= udotn;
        u[2] *= udotn;
		*/

		udotn = DOT_SF16( u, u );
        udotn = NvSFxSqrt( udotn );
		udotn = NvSFxRecip( udotn );
		u[0] = NvSFxMul( udotn, u[0] );
		u[1] = NvSFxMul( udotn, u[1] );
		u[2] = NvSFxMul( udotn, u[2] );

        /* compute the reflection vector */
		/*
        udotn = u[0]*n[0] + u[1]*n[1] + u[2]*n[2];
        r[0] = u[0] - 2.0f*udotn*n[0]; 
        r[1] = u[1] - 2.0f*udotn*n[1]; 
        r[2] = u[2] - 2.0f*udotn*n[2];

        r[2] = r[2]+1;
        m = 2.0f * sqrt( float(r[0]*r[0] + r[1]*r[1] + r[2]*r[2]) );
		*/

		udotn = DOT_SF16( u, n );
		udotn = NvSFxMul( 0x00020000, udotn );
		r[0] = NvSFxSub( u[0], NvSFxMul( udotn, n[0] ) );
		r[1] = NvSFxSub( u[1], NvSFxMul( udotn, n[1] ) );
		r[2] = NvSFxSub( u[2], NvSFxMul( udotn, n[2] ) );

        /* Must normalize (to get correct z-component). */
		udotn = DOT_SF16( r, r );
        udotn = NvSFxSqrt( udotn );
		udotn = NvSFxRecip( udotn );
		r[0] = NvSFxMul( udotn, r[0] );
		r[1] = NvSFxMul( udotn, r[1] );
		r[2] = NvSFxMul( udotn, r[2] );

        /* Compute texture coordinates for texture A.
        /*
        m = (1.0-r[2]);
        tc[0] = 0.5 * (r[0] / m) + 0.5;
        tc[1] = 0.5 * (r[1] / m) + 0.5;
		*/
        tc = shape->vertices[i].tcmapA;
        m = NvSFxSub( 0x00010000, r[2] );
        if ( m == 0x00000000 )
        {
            tc[0] = 0x00008000;
            tc[1] = 0x00008000;
        }
        else
        {
            m = NvSFxRecip( m );
            tc[0] = NvSFxMul( r[0], m );
            tc[1] = NvSFxMul( r[1], m );
            tc[0] = NvSFxMul( 0x00008000, tc[0] );
            tc[0] = NvSFxAdd( 0x00008000, tc[0] );
            tc[1] = NvSFxMul( 0x00008000, tc[1] );
            tc[1] = NvSFxAdd( 0x00008000, tc[1] );
        }

        /* Alpha value should be the 'z' component of the reflection vector */
        /* in with [-1,1] mapped to the range [0,1].                        */
        shape->vertices[i].alpha = (NvSFxAdd( 0x00008000, NvSFxMul( 0x00008000, r[2] ) ) << 16) & 0xFF000000;

        /* Compute texture coordinates for texture B. */
        /*
        m = (r[2]+1.0);
        tc[0] = 0.5 * (-r[0] / m) + 0.5;
        tc[1] = 0.5 * (-r[1] / m) + 0.5;
		*/
        tc = shape->vertices[i].tcmapB;
        m = NvSFxAdd( 0x00010000, r[2] );
        if ( m == 0x00000000 )
        {
            tc[0] = 0x00008000;
            tc[1] = 0x00008000;
        }
        else
        {
            m = NvSFxRecip( m );
            m = NvSFxSub( 0x00000000, m );
            tc[0] = NvSFxMul( r[0], m );
            tc[1] = NvSFxMul( r[1], m );
            tc[0] = NvSFxMul( 0x00008000, tc[0] );
            tc[0] = NvSFxAdd( 0x00008000, tc[0] );
            tc[1] = NvSFxMul( 0x00008000, tc[1] );
            tc[1] = NvSFxAdd( 0x00008000, tc[1] );
        }
    }
}


// Used to hold 1.0f/n values
#define ONE_OVER_SIZE 257
static NvSFx OneOver[ONE_OVER_SIZE];

static void MakeNewVertex( Shape *shape, float latAngle, float longAngle )
{
    Vertex *vert;
    NvSFx cosLat;

    vert = shape->vertices + shape->numVerts++;
    cosLat = NvSFxFloat2Fixed( (float)cos(latAngle) ),
    vert->p[0] = NvSFxMul( cosLat, NvSFxFloat2Fixed( (float)cos(longAngle) ) ),
    vert->p[1] = NvSFxMul( cosLat, NvSFxFloat2Fixed( (float)sin(longAngle) ) ),
    vert->p[2] = NvSFxFloat2Fixed( (float)sin(latAngle) );

    // set home position
    vert->h[0] = vert->p[0];
    vert->h[1] = vert->p[1];
    vert->h[2] = vert->p[2];

    // zero initial velocity
    vert->v[0] =
    vert->v[1] =
    vert->v[2] = 0x00000000; /* 0.0f */

    // zero average velocity
    vert->a[0] =
    vert->a[1] =
    vert->a[2] = 0x00000000; /* 0.0f */
}

static void MakeNewEdge( Shape *shape, int vertId0, int vertId1 )
{
    Vertex *v0, *v1;
    Edge *edge;
//    fix dx, dy, dz;
    NvSFx dx, dy, dz;

    v0 = shape->vertices + vertId0,
    v1 = shape->vertices + vertId1;
    edge = shape->edges + shape->numEdges++;
    edge->v0id = vertId0;
    edge->v1id = vertId1;

//    dx = v1->p[0] - v0->p[0];
//    dy = v1->p[1] - v0->p[1];
//    dz = v1->p[2] - v0->p[2];
//    edge->l = sqrt(float(dx*dx + dy*dy + dz*dz));
    dx = NvSFxSub( v1->p[0], v0->p[0] );
    dy = NvSFxSub( v1->p[1], v0->p[1] );
    dz = NvSFxSub( v1->p[2], v0->p[2] );
    edge->l = NvSFxSqrt( NvSFxAdd( NvSFxMul(dx,dx), NvSFxAdd( NvSFxMul(dy,dy), NvSFxMul(dz,dz) ) ) );
}


void ShapeDestroy( Shape *shape )
{
    int i;

    // check to see if we need to free memory
    if ( shape->vertices )
    {
        free( shape->vertices );
        shape->vertices = NULL;
    }
    if ( shape->edges )
    {
        free( shape->edges );
        shape->edges = NULL;
    }
    if ( shape->tristrips )
    {
        for ( i = 0; i < shape->numTristrips; i++ )
        {
            free( shape->tristrips[i].vertices );
            free( shape->tristrips[i].indices );
        }
        free( shape->tristrips );
        shape->tristrips = NULL;
    }

    free( shape );
    shape = NULL;
}

Shape* ShapeBuildSphere( int depth )
{

#define EXPECTED_VERTS      ( 4*depth*depth + 2)
#define EXPECTED_EDGES      (12*depth*depth)
#define NUM_TRISTRIPS       ( 4*depth)
#define NUM_VERTS_PER_STRIP ( 2*depth + 2)

    Shape *shape ;
    int i, j, quadrant, vertId;
    int vertTab[MAX_DEPTH+1][MAX_DEPTH+1];
    float oo_depth = 1.0f/(float)depth;

    shape = (Shape*) calloc(1, sizeof(Shape));
    ShapeInitialize( shape );

    // check that we're not overly ambitious
    if ( depth > MAX_DEPTH )
    {
        printf("Increase MAX_DEPTH\n");
        return NULL;
    }
    if ( 4*depth > ONE_OVER_SIZE-1 )
    {
        printf("Increase ONE_OVER_SIZE\n");
        return NULL;
    }

    // initialize OneOver table
    OneOver[0] = 0x00000000; /* 0.0f */
    for ( i = 1; i < ONE_OVER_SIZE; i++ )
    {
        OneOver[i] = NvSFxFloat2Fixed( 1.0f/(float)i );
    }

    // allocate all the space we'll use up front
    shape->numVerts = 0;
    shape->numEdges = 0;
    shape->numTristrips = NUM_TRISTRIPS;
    shape->vertices  = (Vertex*)   calloc(EXPECTED_VERTS, sizeof(Vertex));
    shape->edges     = (Edge*)     calloc(EXPECTED_EDGES, sizeof(Edge));
    shape->tristrips = (Tristrip*) calloc(NUM_TRISTRIPS,  sizeof(Tristrip));
    for ( i=0 ; i<NUM_TRISTRIPS ; i++ ) {
        shape->tristrips[i].numVerts = NUM_VERTS_PER_STRIP;
        shape->tristrips[i].vertices =
          (Vertex**)        calloc(NUM_VERTS_PER_STRIP, sizeof(Vertex*));
        shape->tristrips[i].indices  =
          (unsigned short*) calloc(NUM_VERTS_PER_STRIP, sizeof(unsigned short));
    }

    // generate all the vertices
    MakeNewVertex(shape, (float)(0.5f*M_PI), 0.0f);
    for ( i=1 ; i<2*depth ; i++ ) {
        int numCuts = (i<=depth) ? 4*i : 4*(2*depth-i);
        float latAngle = (float)(0.5f*M_PI*(1.0f - (float)i*oo_depth));
        for ( j=0 ; j<numCuts ; j++ ) {
            float longAngle = (float)(2.0f * M_PI * (float)j * (1.0f/numCuts));
            MakeNewVertex(shape, latAngle, longAngle);
        }
    }
    MakeNewVertex(shape, (float)(-0.5f*M_PI), 0.0f);

    if (shape->numVerts != EXPECTED_VERTS) {
        printf("Whoops, wrong number of verts (%d observed, %d expected)\n",
                shape->numVerts, EXPECTED_VERTS);
    }

    for ( quadrant=0 ; quadrant<4 ; quadrant++ ) {

        // generate vertTab, a table of vertex indices used to build
        // edges and tristrips
        vertTab[0][0] = 0;
        vertTab[depth][depth] = shape->numVerts-1;
        for ( i=1, vertId=1 ; i<2*depth ; i++ ) {
            int numCuts, startCol, startRow;
            if (i <= depth) {
                numCuts  = 4*i;
                startCol = i;
                startRow = 0;
            } else {
                numCuts  = 4*(2*depth-i);
                startCol = depth;
                startRow = i-depth;
            }
            for ( j=0 ; j<=(numCuts>>2) ; j++ ) {
                if ((quadrant == 3) &&
                        ((startCol-j == 0) || (startRow+j == depth))) {
                    // the 4th quadrant is special because it has to wrap
                    // back around to the 1st quadrant
                    vertTab[startCol-j][startRow+j] =
                        vertId + j + quadrant*(numCuts>>2) - numCuts;
                } else {
                    vertTab[startCol-j][startRow+j] =
                        vertId + j + quadrant*(numCuts>>2);
                }
            }
            vertId += numCuts;
        }

        // generate edges
        // "forward slash" edges
        for ( i=1 ; i<=depth ; i++ ) {
            for ( j=0 ; j<depth ; j++ ) {
                MakeNewEdge(shape, vertTab[i][j], vertTab[i][j+1]);
            }
        }
        // "backward slash" edges
        for ( j=0 ; j<depth ; j++ ) {
            for ( i=0 ; i<depth ; i++ ) {
                MakeNewEdge(shape, vertTab[i][j], vertTab[i+1][j]);
            }
        }
        // "horizontal" edges
        for ( i=1 ; i<2*depth ; i++ ) {
            int rowEdges, startCol, startRow;
            if (i <= depth) {
                rowEdges = i;
                startCol = i;
                startRow = 0;
            } else {
                rowEdges = 2*depth-i;
                startCol = depth;
                startRow = i-depth;
            }
            for ( j=0 ; j<rowEdges ; j++ ) {
                MakeNewEdge(shape, vertTab[startCol-j]  [startRow+j],
                                  vertTab[startCol-j-1][startRow+j+1]);
            }
        }

        // generate tristrips
        {
        Tristrip *tstrip = shape->tristrips + quadrant * depth;
        for ( j=0 ; j<depth ; j++, tstrip++ ) {

            for ( i=0 ; i<=depth ; i++ ) {

                tstrip->indices[2*i+0] = (unsigned short)vertTab[i][j];
                tstrip->indices[2*i+1] = (unsigned short)vertTab[i][j+1];

                tstrip->vertices[2*i+0] = shape->vertices + vertTab[i][j];
                tstrip->vertices[2*i+1] = shape->vertices + vertTab[i][j+1];
            }

            for ( i=0 ; i<tstrip->numVerts-2 ; i++ ) {

                // bump the connectivity count
                tstrip->vertices[i+0]->connectedness++;
                tstrip->vertices[i+1]->connectedness++;
                tstrip->vertices[i+2]->connectedness++;
            }
        }
        }
    }

    if (shape->numEdges != EXPECTED_EDGES) {
        printf("Whoops, wrong number of edges (%d observed, %d expected)\n",
                shape->numEdges, EXPECTED_EDGES);
    }

    return shape;
}


void ShapeCalcVelocity( Shape *shape )
{
    Vertex *vert;
    Edge *edge;
    int i;
    NvSFx k1 = NvSFxFloat2Fixed( 0.005f ); // spring home
    NvSFx k2 = NvSFxFloat2Fixed( 0.6f );   // edge spring
    NvSFx half_oo_nc;
    NvSFx fx,  fy,  fz;
    NvSFx fLen, scale;
    NvSFx v0scale, v1scale;

    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
        /* fix half_oo_nc = 0.5f * OneOver[vert->connectedness]; */
       	half_oo_nc = NvSFxMul( 0x00008000, OneOver[vert->connectedness] );

	/*
        vert->v[0] += (vert->h[0] - vert->p[0]) * k1 + vert->n[0] * half_oo_nc;
        vert->v[1] += (vert->h[1] - vert->p[1]) * k1 + vert->n[1] * half_oo_nc;
        vert->v[2] += (vert->h[2] - vert->p[2]) * k1 + vert->n[2] * half_oo_nc;
	*/
        vert->v[0] =
			NvSFxAdd( vert->v[0],
					  NvSFxAdd(
						  NvSFxMul(
                                NvSFxSub( vert->h[0], vert->p[0] ),
                                k1
                          ),
                          NvSFxMul( vert->n[0], half_oo_nc )
			          )
            );

        vert->v[1] =
			NvSFxAdd( vert->v[1],
					  NvSFxAdd(
						  NvSFxMul(
                                NvSFxSub( vert->h[1], vert->p[1] ),
                                k1
                          ),
                          NvSFxMul( vert->n[1], half_oo_nc )
			          )
            );

        vert->v[2] =
			NvSFxAdd( vert->v[2],
					  NvSFxAdd(
						  NvSFxMul(
                                NvSFxSub( vert->h[2], vert->p[2] ),
                                k1
                          ),
                          NvSFxMul( vert->n[2], half_oo_nc )
			          )
            );

        vert->a[0] =
        vert->a[1] =
        vert->a[2] =
        vert->n[0] =
        vert->n[1] =
        vert->n[2] = 0x00000000; /* initialize to 0.0 */
    }

    for ( i = 0, edge = shape->edges; i < shape->numEdges; i++, edge++ )
    {
        Vertex *vert0 = shape->vertices + edge->v0id;
        Vertex *vert1 = shape->vertices + edge->v1id;

		/*
        fx = vert1->p[0] - vert0->p[0];
        fy = vert1->p[1] - vert0->p[1];
        fz = vert1->p[2] - vert0->p[2];
		*/

        fx = NvSFxSub( vert1->p[0], vert0->p[0] );
        fy = NvSFxSub( vert1->p[1], vert0->p[1] );
        fz = NvSFxSub( vert1->p[2], vert0->p[2] );

		/*
        fLen = sqrt(float(fx*fx + fy*fy + fz*fz));
		*/

        fLen = NvSFxAdd(
					NvSFxAdd(
						NvSFxMul(fx,fx), NvSFxMul(fy,fy) ),
					NvSFxMul(fz,fz)
               );
        fLen = NvSFxSqrt(fLen);

		// Tests for perf comparisons.
        //fLen = NvSFxSqrt((fx*fx + fy*fy + fz*fz));
        //fLen = 0x00010000;
		// End tests.
		

        /* ensure we don't divide by zero. */
		/*
        scale = (fLen - edge->l);
        if ( scale < 0 )
            scale = 0;
        else
            scale = k2 * scale / fLen;
		*/

        scale = NvSFxSub( fLen,  edge->l );
        if ( scale & 0x80000000 )
            scale = 0x00000000;
        else
		{
            scale = NvSFxMul( k2, scale );
            scale = NvSFxDiv( scale, fLen );
		}
#if 0
        if ( fLen == 0 ) scale.v = 0x00000000;
        else
        scale = k2 * (fLen - edge->l) / fLen;
#endif
#if 0
        if ( fLen == 0 ) fLen.v = 0x00000001;

        scale = k2 * (fLen - edge->l) / fLen;
#endif

        // Check for vertices connected to 4 edges.  Since all but
        // 6 verts are connected to 6 edges, scale up the weight
        // of this edge to make if effectively the same as if the
        // vertex was connected to 6 edges.
        /*
        v1scale = (vert1->connectedness == 4) ? 1.5f * scale : scale;
        v0scale = (vert0->connectedness == 4) ? 1.5f * scale : scale;
        */
        v1scale = (vert1->connectedness == 4) ? NvSFxMul( 0x00018000, scale) : scale;
        v0scale = (vert0->connectedness == 4) ? NvSFxMul( 0x00018000, scale) : scale;

		/*
        vert0->a[0] += (vert1->v[0] -= v1scale * fx);
        vert0->a[1] += (vert1->v[1] -= v1scale * fy);
        vert0->a[2] += (vert1->v[2] -= v1scale * fz);

        vert1->a[0] += (vert0->v[0] += v0scale * fx);
        vert1->a[1] += (vert0->v[1] += v0scale * fy);
        vert1->a[2] += (vert0->v[2] += v0scale * fz);
		*/

        vert1->v[0] = NvSFxSub( vert1->v[0], NvSFxMul( v1scale, fx ) );
		vert0->a[0] = NvSFxAdd( vert0->a[0], vert1->v[0] );
        vert1->v[1] = NvSFxSub( vert1->v[1], NvSFxMul( v1scale, fy ) );
		vert0->a[1] = NvSFxAdd( vert0->a[1], vert1->v[1] );
        vert1->v[2] = NvSFxSub( vert1->v[2], NvSFxMul( v1scale, fz ) );
		vert0->a[2] = NvSFxAdd( vert0->a[2], vert1->v[2] );

        vert0->v[0] = NvSFxAdd( vert0->v[0], NvSFxMul( v0scale, fx ) );
		vert1->a[0] = NvSFxAdd( vert1->a[0], vert0->v[0] );
        vert0->v[1] = NvSFxAdd( vert0->v[1], NvSFxMul( v0scale, fy ) );
		vert1->a[1] = NvSFxAdd( vert1->a[1], vert0->v[1] );
        vert0->v[2] = NvSFxAdd( vert0->v[2], NvSFxMul( v0scale, fz ) );
		vert1->a[2] = NvSFxAdd( vert1->a[2], vert0->v[2] );
    }
}

void ShapeFilterVelocity( Shape *shape )
{
    Vertex *vert;
    int i;
    NvSFx tenth_oo_c;
    NvSFx v0, v1, v2;


	/*
    shape->drag += 0.01f;
    if ( shape->drag > shape->final_drag )
        shape->drag = shape->final_drag;

    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
        fix tenth_oo_c = 0.1f * OneOver[vert->connectedness];
        fix v0, v1, v2;

        v0 = (vert->v[0] * 0.9f) + (vert->a[0] * tenth_oo_c);
        v1 = (vert->v[1] * 0.9f) + (vert->a[1] * tenth_oo_c);
        v2 = (vert->v[2] * 0.9f) + (vert->a[2] * tenth_oo_c);

        vert->p[0] += v0;
        vert->p[1] += v1;
        vert->p[2] += v2;

        vert->v[0] = v0 * shape->drag;
        vert->v[1] = v1 * shape->drag;
        vert->v[2] = v2 * shape->drag;
    }
	*/

	/* Assumption is that drag is always positive. */
	/* Assumption is that final_drag is always positive. */
	if ( shape->drag != shape->final_drag )
	{
        shape->drag = NvSFxAdd( shape->drag, 0x0000028F );
        if ( shape->drag > shape->final_drag )
            shape->drag = shape->final_drag;
	}

    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
        tenth_oo_c = NvSFxMul( 0x0000199A,  OneOver[vert->connectedness] );

		v0 = NvSFxAdd( NvSFxMul( vert->v[0], 0x0000E666 ),
				       NvSFxMul( vert->a[0], tenth_oo_c ) );
		v1 = NvSFxAdd( NvSFxMul( vert->v[1], 0x0000E666 ),
				       NvSFxMul( vert->a[1], tenth_oo_c ) );
		v2 = NvSFxAdd( NvSFxMul( vert->v[2], 0x0000E666 ),
				       NvSFxMul( vert->a[2], tenth_oo_c ) );

		vert->p[0] = NvSFxAdd( vert->p[0], v0 );
		vert->p[1] = NvSFxAdd( vert->p[1], v1 );
		vert->p[2] = NvSFxAdd( vert->p[2], v2 );

		vert->v[0] = NvSFxMul( v0, shape->drag );
		vert->v[1] = NvSFxMul( v1, shape->drag );
		vert->v[2] = NvSFxMul( v2, shape->drag );
    }

}


// e is the eye point, n is the line of sight, t is the test point
// returns distance squared
NvSFx ShapePickDistance( NvSFx e[3], NvSFx n[3], NvSFx t[3] )
{
	/*
    fix s;
    fix temp[3];
    temp[0] = t[0]-e[0];
    temp[1] = t[1]-e[1];
    temp[2] = t[2]-e[2];
    s = dot(n,temp)/dot(n,n);
    temp[0] = e[0] + n[0]*s;
    temp[1] = e[1] + n[1]*s;
    temp[2] = e[2] + n[2]*s;

    temp[0] = t[0]-temp[0];
    temp[1] = t[1]-temp[1];
    temp[2] = t[2]-temp[2];

    s = dot(temp,temp);
    return s;
	*/

    NvSFx s;
    NvSFx temp[3];
    temp[0] = NvSFxSub( t[0], e[0] );
    temp[1] = NvSFxSub( t[1], e[1] );
    temp[2] = NvSFxSub( t[2], e[2] );
	s = DOT_SF16( n, n );
	s = NvSFxRecip( s );
	s = NvSFxMul( s, DOT_SF16( n, temp ) );
    temp[0] = NvSFxAdd( e[0], NvSFxMul( n[0], s ) );
    temp[1] = NvSFxAdd( e[1], NvSFxMul( n[1], s ) );
    temp[2] = NvSFxAdd( e[2], NvSFxMul( n[2], s ) );

    temp[0] = NvSFxSub( t[0], temp[0] );
    temp[1] = NvSFxSub( t[1], temp[1] );
    temp[2] = NvSFxSub( t[2], temp[2] );

    s = DOT_SF16( temp, temp );
    return s;
}


/*
 * powit10(p) -- computes 10^p within 0.015% error.
 * 
 *      logA(x) = logB(x) / logB(A)
 *
 *         10^p = 2^log2(10^p)
 *              = 2^(p*log2(10))
 *              = 2^A    where A = p * log10(10)/log10(2)
 *              = 2^A    where A = p / log10(2)
 *              = 2^(Int(A)+Frac(A))
 *              = (2^Int(A))*(2^Frac(A))
 *
 */

static int log2_10 = 0x0003526A;
static int two_exp_half_i[16] = { 0x00016A0A, 0x00013070, 0x0001172C, 0x00010B56,
                                  0x0001059B, 0x000102CA, 0x00010164, 0x000100B2,
                                  0x00010059, 0x0001002C, 0x00010016, 0x0001000B,
                                  0x00010006, 0x00010003, 0x00010001, 0x00010001 };
static int half_exp_half_i[16] = { 0x0000B505, 0x0000D745, 0x0000EAC1, 0x0000F525,
                                   0x0000FA84, 0x0000FD3E, 0x0000FE9E, 0x0000FF4F,
                                   0x0000FFA7, 0x0000FFD4, 0x0000FFEA, 0x0000FFF5,
                                   0x0000FFFA, 0x0000FFFD, 0x0000FFFF, 0x0000FFFF };

static int powit10( NvSFx p )
{
    int A;
    int result;
    int intA;
    int bitmaski;
    int i;

    A = NvSFxMul( p, log2_10 );

    intA = A >> 16;
    result = (0x10000 << intA);    /* 2^Int(A) */
    for ( i = 0; i < 16; i++ )
    {
        bitmaski = (0x00008000 >> i);
        if ( A & bitmaski )
            result = NvSFxMul( result, two_exp_half_i[i] );
    }

	/*
	fix temp;
	temp = log2_10;
	fprintf( stderr, "log2_10 = %.8X\n", temp.v );
	for ( i = 0; i < 16; i++ )
	{
	    temp = pow( 2.0, pow( 0.5, i+1 ) );
	    fprintf( stderr, "2^pow( 0.5, %d ) = %.8X\n", i+1, temp.v );
	}
	fprintf( stderr, "\n" );
	*/

    return result;
}


/*
 * invpowit10(p) -- computes 10^(-p) within 0x0000000b maximum absolute error.
 *                                   within 2^-13 + 2^-15 maximum absolute error.
 *                                   within 0.00015259 maximum absolute error.
 *                                   
 * In practice, the error is not too large.
 * If more accuracy is needed use NvSFxRecip( powit10(p) ) with slight degradation
 * in performance.
 * 
 *      logA(x) = logB(x) / logB(A)
 *
 *         10^(-p) = 2^log2(10^(-p))
 *                 = 2^(-p*log2(10))
 *                 = 2^(-A)    where A = p * log10(10)/log10(2)
 *                 = 2^(-A)    where A = p / log10(2)
 *                 = 0.5^(Int(A)+Frac(A))
 *                 = (0.5^Int(A))*(0.5^Frac(A))
 *
 */

static int invpowit10( NvSFx p )
{
    int A;
    int result;
    int intA;
    int bitmaski;
    int i;

    A = NvSFxMul( p, log2_10 );

    intA = A >> 16;
    result = (0x00010000 >> intA);  /* 0.5^Int(A) */
    for ( i = 0; i < 16; i++ )
    {
        bitmaski = (0x00008000 >> i);
        if ( A & bitmaski )
            result = NvSFxMul( result, half_exp_half_i[i] );
    }

	/*
	fix temp;
	temp = log2_10;
	fprintf( stderr, "log2_10 = %.8X\n", temp.v );
	for ( i = 0; i < 16; i++ )
	{
	    temp = pow( 0.5, pow( 0.5, i+1 ) );
	    fprintf( stderr, "0.5^pow( 0.5, %d ) = %.8X\n", i+1, temp.v );
	}
	fprintf( stderr, "\n" );
	*/

    return result;
}


void ShapePick( Shape *shape, NvSFx e[3], NvSFx n[3] )
{
    NvSFx closest_distance;
    Vertex *vert, *closest;
    NvSFx *p0;
    int i;
    NvSFx distance;
    NvSFx *t;
    NvSFx *vn;

    /* Used for the 2nd for loop. */
    NvSFx s;
    NvSFx *p1;
    NvSFx v[3];

    /* Initialize to the largest value. */
	closest_distance = 0x7FFFFFFF;
    closest = NULL;

    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
		/*
        fix distance;
        fix t[3];
        fix vn[3];
        t[0] = vert->p[0];
        t[1] = vert->p[1];
        t[2] = vert->p[2];
        vn[0] = vert->n[0];
        vn[1] = vert->n[1];
        vn[2] = vert->n[2];

        distance = ShapePickDistance( e, n, t );
		*/

        t  = vert->p;
        vn = vert->n;

        distance = ShapePickDistance( e, n, t );

		/*
        if ( (distance < closest_distance) && (dot(vn, n)<0.0f) )
		*/
        if ( (distance < closest_distance) && (DOT_SF16(vn, n)&0x8000000) )
        {
            closest_distance = distance;
            closest = vert;
        }
    }

    if (!closest)
        return;

    p0 = closest->p;

    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
		/*
        fix s;
        fix *p1 = vert->p;
        fix v[3];
        v[0] = p1[0] - p0[0];
        v[1] = p1[1] - p0[1];
        v[2] = p1[2] - p0[2];

        s = pow(10.0f, float(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
        s = 1.0f/s;

        vert->v[0] -= vert->p[0] * s * 0.15f;
        vert->v[1] -= vert->p[1] * s * 0.15f;
        vert->v[2] -= vert->p[2] * s * 0.15f;
		*/
		p1 = vert->p;
        v[0] = NvSFxSub( p1[0], p0[0] );
        v[1] = NvSFxSub( p1[1], p0[1] );
        v[2] = NvSFxSub( p1[2], p0[2] );

        s = DOT_SF16( v, v );
		s = invpowit10( s );
		//s.v = NvSFxRecip( powit10( s ) );
		//s = 1.0f / pow10( float(s) );
		s = NvSFxMul( s, 0x00002666 );
        vert->v[0] = NvSFxSub( vert->v[0], NvSFxMul( vert->p[0], s ) );
        vert->v[1] = NvSFxSub( vert->v[1], NvSFxMul( vert->p[1], s ) );
        vert->v[2] = NvSFxSub( vert->v[2], NvSFxMul( vert->p[2], s ) );
    }
}

void ShapeDump( Shape *shape )
{
    Vertex *vert;
    Edge *edge;
    Tristrip *tstrip;
    int i, j;

    printf("numVerts %d\n", shape->numVerts);
    for ( i = 0, vert = shape->vertices; i < shape->numVerts; i++, vert++ )
    {
        printf("  %d 0x%p: 0x%x 0x%x 0x%x\n", i, (void *) vert,
                vert->p[0], vert->p[1], vert->p[2] );
    }

    printf("numEdges %d\n", shape->numEdges);
    for ( i = 0, edge = shape->edges; i < shape->numEdges; i++, edge++ )
    {
        printf("  %d 0x%p: %d %d\n", i, (void *) edge,
                (int)edge->v0id, (int)edge->v1id );
    }

    printf("numTristrip %d\n", shape->numTristrips);
    for ( i = 0, tstrip = shape->tristrips; i < shape->numTristrips; i++, tstrip++ )
    {
        printf(" ");
        for ( j = 0; j < tstrip->numVerts; j++ )
        {
            printf(" 0x%p", (void *) tstrip->vertices[j]);
        }
        printf("\n");
    }
}

static float quad_verticesA[2*4] =
{
    0.0f,  1.0f,
   -2.0f,  1.0f,
    0.0f, -1.0f,        
   -2.0f, -1.0f
};
static float quad_verticesB[2*4] =
{
    2.0f,  1.0f,
    0.0f,  1.0f,
    2.0f, -1.0f,        
    0.0f, -1.0f
};

static float quad_texcoords[2*4] =
{  
    1.0f,  1.0f,
    0.0f,  1.0f,
    1.0f,  0.0f,       
    0.0f,  0.0f 
};

void ShapeDrawSphereMapTexture()
{
    glEnableClientState ( GL_VERTEX_ARRAY );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );
    glEnable( GL_TEXTURE_2D );

    glVertexPointer( 2, GL_FLOAT, 0, quad_verticesA );
    glTexCoordPointer( 2, GL_FLOAT, 0, quad_texcoords );
    glBindTexture( GL_TEXTURE_2D, texidA[current] );
    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

    glVertexPointer( 2, GL_FLOAT, 0, quad_verticesB );
    glTexCoordPointer( 2, GL_FLOAT, 0, quad_texcoords );
    glBindTexture( GL_TEXTURE_2D, texidB[current] );
    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

    glDisable( GL_TEXTURE_2D );
    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisableClientState ( GL_VERTEX_ARRAY );
}
