/*
 *
  Copyright 2003, NVIDIA Corporation.
  All rights Reserved.

  THE INFORMATION CONTAINED HEREIN IS PROPRIETARY AND CONFIDENTIAL TO
  NVIDIA, CORPORATION.  USE, REPRODUCTION OR DISCLOSURE TO ANY THIRD PARTY
  IS SUBJECT TO WRITTEN PRE-APPROVAL BY NVIDIA, CORPORATION.
 *
 * Simple app that draws a triangle using fixed-point entry points
 * into the driver.
 */

#define GLUT_NO_WARNING_DISABLE
#include "GL\glutes.h"

#define FIXINT(x) (x << 16)
#define FIXHALF (FIXINT(1) >> 1)

#define ROOT_3_OVER_2 ((GLfixed) (unsigned) (0.8660254 * (1 << 16)))
#define ROOT_3_OVER_6 (ROOT_3_OVER_2/3)

static GLubyte ind [3] = { 0, 1, 2 };
static GLfixed vert[6] = { FIXHALF, -ROOT_3_OVER_6,
                          -FIXHALF, -ROOT_3_OVER_6,
                           FIXINT(0),  ROOT_3_OVER_2 - ROOT_3_OVER_6 };
static GLfixed tex [6] = { FIXINT(1), FIXINT(0),
                           FIXINT(0), FIXINT(0),
                           FIXHALF, FIXINT(1) };
static GLfixed col [9] = { FIXINT(1), FIXINT(0), FIXINT(0),
                           FIXINT(0), FIXINT(1), FIXINT(0),
                           FIXINT(0), FIXINT(0), FIXINT(1) };
static GLfixed angle = 0;
static int animate = 0;

static void InitGraphicsState()
{
    glClearColorx(0, 0, 0, 0);
    glOrthox(-FIXINT(1), FIXINT(1), -FIXINT(1),
             FIXINT(1), -FIXINT(1), FIXINT(1));
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glVertexPointer  (2, GL_FIXED, 0, (void *) vert);
    glTexCoordPointer(2, GL_FIXED, 0, (void *) tex);
    glColorPointer   (3, GL_FIXED, 0, (void *) col);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

static void Display(void)
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glLoadIdentity();
    glRotatex(angle, 0, 0, FIXINT(1));
    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_BYTE, ind);
    glutSwapBuffers();
}

static void idle(void)
{
    if ( animate )
        angle += FIXINT(15) / 2000;

    glutPostRedisplay();
}

static void keyboard(unsigned char key, int x, int y)
{
    switch (key) {
        case ' ':
            animate = !animate;
            break;
        case 27:
            exit(0);
    }
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowPosition(200, 200);
    glutInitWindowSize(320, 240);
    glutCreateWindow("hello");
    
    InitGraphicsState();

    glutDisplayFunc(Display);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    glutMainLoop();

    exit(0);
}
